@@ -1,7 +1,7 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2014 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2014-2016 -- leonerd@leonerd.org.uk
 
 package Net::Async::Matrix;
 
@@ -11,25 +11,27 @@ use warnings;
 use base qw( IO::Async::Notifier );
 IO::Async::Notifier->VERSION( '0.63' ); # adopt_future
 
-our $VERSION = '0.11_002';
+our $VERSION = '0.18_001';
 $VERSION = eval $VERSION;
 
 use Carp;
 
 use Future;
 use Future::Utils qw( repeat );
-use JSON qw( encode_json decode_json );
+use JSON::MaybeXS qw( encode_json decode_json );
 
 use Data::Dump 'pp';
+use File::stat;
 use Struct::Dumb;
 use Time::HiRes qw( time );
+use URI;
 
 struct User => [qw( user_id displayname presence last_active )];
 
 use Net::Async::Matrix::Room;
 
-use constant PATH_PREFIX => "/_matrix/client/api/v1";
-use constant LONGPOLL_SECONDS => 30;
+use constant PATH_PREFIX => "/_matrix/client/r0";
+use constant LONGPOLL_TIMEOUT => 30;
 
 # This is only needed for the (undocumented) recaptcha bypass feature
 use constant HAVE_DIGEST_HMAC_SHA1 => eval { require Digest::HMAC_SHA1; };
@@ -52,7 +54,7 @@ C<Net::Async::Matrix> - use Matrix with L<IO::Async>
  $loop->add( $matrix );
 
  $matrix->login(
-    user     => '@my-user:home.server',
+    user_id  => '@my-user:home.server',
     password => 'SeKr1t',
  )->get;
 
@@ -141,6 +143,21 @@ this should not need modification.
 
 Optional. Sets default event handlers on new room objects.
 
+=head2 enable_events => BOOL
+
+Optional. Normally enabled, but if set to a defined-but-false value (i.e. 0 or
+empty string) the event stream will be disabled. This will cause none of the
+incoming event handlers to be invoked, because the server will not be polled
+for events.
+
+This may be useful in simple send-only cases where the client has no interest
+in receiveing any events, and wishes to reduce the load on the homeserver.
+
+=head2 longpoll_timeout => NUM
+
+Optional. Timeout in seconds for the C</events> longpoll operation. Defaults
+to 30 seconds if not supplied.
+
 =cut
 
 sub _init
@@ -157,17 +174,27 @@ sub _init
          fail_on_error => 1,
          max_connections_per_host => 3, # allow 2 longpolls + 1 actual command
          user_agent => __PACKAGE__,
+         pipeline => 0,
       );
       $self->add_child( $ua );
       $ua
    };
 
+   # Injectable for unit tests, other event systems, etc..
+   # For now undocumented while I try to work out the wider design issues
+   $self->{make_delay} = delete $params->{make_delay} || $self->_capture_weakself( sub {
+      my ( $self, $secs ) = @_;
+      $self->loop->delay_future( after => $secs );
+   } );
+
    $self->{msgid_next} = 0;
 
    $self->{users_by_id} = {};
    $self->{rooms_by_id} = {};
 
    $self->{path_prefix} = PATH_PREFIX;
+
+   $self->{longpoll_timeout} = LONGPOLL_TIMEOUT;
 }
 
 =head1 METHODS
@@ -182,7 +209,7 @@ sub configure
    my $self = shift;
    my %params = @_;
 
-   foreach (qw( server path_prefix ua SSL
+   foreach (qw( server path_prefix ua SSL enable_events longpoll_timeout
                 on_log on_unknown_event on_presence on_room_new on_room_del on_invite
                 on_room_member on_room_message )) {
       $self->{$_} = delete $params{$_} if exists $params{$_};
@@ -288,12 +315,71 @@ sub _do_DELETE
    );
 }
 
-=head2 $matrix->login( %params )->get
+sub _do_POST_file
+{
+   my $self = shift;
+   my ( $path, %params ) = @_;
+
+   my $uri = $self->_uri_for_path( "" );
+   $uri->path( "/_matrix" . $path );
+
+   my $req = HTTP::Request->new( "POST" , $uri );
+   $req->header( Content_type => $params{content_type} );
+
+   my $body;
+
+   if( defined $params{content} ) {
+      $req->content( $params{content} );
+      $req->header( Content_length => length $req->content );
+   }
+   elsif( defined $params{file} or defined $params{fh} ) {
+      my $fh = $params{fh};
+      $fh or open $fh, "<", $params{file} or
+         return Future->fail( "Cannot read $params{file} - $!", open => );
+
+      $body = sub {
+         $fh->read( my $buffer, 65536 ) or return undef;
+         return $buffer;
+      };
+
+      $req->header( Content_length => $params{content_length} // ( stat $fh )->size );
+   }
+
+   my $f = $self->{ua}->do_request(
+      request => $req,
+      request_body => $body,
+   )->then( sub {
+      my ( $response ) = @_;
+      $response->content_type eq "application/json" or
+         return Future->fail( "Expected application/json response", matrix => );
+
+      my $content = $response->content;
+      my $uri;
+      if( length $content and $content ne q("") ) {
+         eval {
+            $content = decode_json( $content );
+            1;
+         } or  
+            return Future->fail( "Unable to parse JSON response " );
+         return Future->done( $content, $response );
+      }
+      else {
+         return Future->done( undef, $response );
+      }
+   });
+
+   return $self->adopt_future( $f );
+}
+
+=head2 login
+
+   $matrix->login( %params )->get
 
 Performs the necessary steps required to authenticate with the configured
-Home Server, actually obtain an access token and starting the event stream.
-The returned C<Future> will eventually yield the C<$matrix> object itself, so
-it can be easily chained.
+Home Server, actually obtain an access token and starting the event stream
+(unless disabled by the C<enable_events> option being false). The returned
+C<Future> will eventually yield the C<$matrix> object itself, so it can be
+easily chained.
 
 There are various methods of logging in supported by Matrix; the following
 sets of arguments determine which is used:
@@ -322,8 +408,10 @@ sub login
 
    if( defined $params{user_id} and defined $params{access_token} ) {
       $self->{$_} = $params{$_} for qw( user_id access_token );
-      $self->start;
-      return Future->done( $self );
+      $self->configure( notifier_name => "uid=$params{user_id}" );
+      return ( ( $self->{enable_events} // 1 ) ? $self->start : Future->done )->then( sub {
+         Future->done( $self )
+      });
    }
 
    # Otherwise; try to obtain the login flow information
@@ -357,12 +445,14 @@ sub _login_with_password
       { type => "m.login.password", user => $params{user_id}, password => $params{password} }
    )->then( sub {
       my ( $resp ) = @_;
-      return $self->login( %$resp ) if defined $resp->{access_token};
+      return $self->login( %$resp, %params ) if defined $resp->{access_token};
       return Future->fail( "Expected server to respond with 'access_token'", matrix => );
    });
 }
 
-=head2 $matrix->register( %params )->get
+=head2 register
+
+   $matrix->register( %params )->get
 
 Performs the necessary steps required to create a new account on the
 configured Home Server.
@@ -460,14 +550,49 @@ sub _register_with_recaptcha
    };
 }
 
-=head2 $f = $matrix->start
+=head2 sync
 
-Performs the initial IMSync on the server, and starts the event stream to
+   $matrix->sync( %params )->get
+
+Performs a single C</sync> request on the server, returning the raw results
+directly.
+
+Takes the following named parameters
+
+=over 4
+
+=item since => STRING
+
+Optional. Sync token from the previous request.
+
+=back
+
+=cut
+
+sub sync
+{
+   my $self = shift;
+   my ( %params ) = @_;
+
+   $self->_do_GET_json( "/sync", %params );
+}
+
+sub await_synced
+{
+   my $self = shift;
+   return $self->{synced_future} //= $self->loop->new_future;
+}
+
+=head2 start
+
+   $f = $matrix->start
+
+Performs the initial sync on the server, and starts the event stream to
 begin receiving events.
 
 While this method does return a C<Future> it is not required that the caller
 keep track of this; the object itself will store it. It will complete when the
-initial IMSync has fininshed, and the event stream has started.
+initial sync has fininshed, and the event stream has started.
 
 If the initial sync has already been requested, this method simply returns the
 future it returned the last time, ensuring that you can await the client
@@ -482,43 +607,27 @@ sub start
    defined $self->{access_token} or croak "Cannot ->start without an access token";
 
    return $self->{start_f} ||= do {
-      my $f = $self->_do_GET_json( "/initialSync", limit => 0 )
-      ->then( sub {
+      undef $self->{synced_future};
+
+      foreach my $room ( values %{ $self->{rooms_by_id} } ) {
+         $room->_reset_for_sync;
+      }
+
+      $self->sync->then( sub {
          my ( $sync ) = @_;
 
-         foreach ( @{ $sync->{rooms} } ) {
-            my $room_id = $_->{room_id};
-            my $membership = $_->{membership};
+         $self->_incoming_sync( $sync );
 
-            if( $membership eq "join" ) {
-               my $state = $_->{state};
+         $self->start_longpoll( since => $sync->{next_batch} );
 
-               my $room = $self->_make_room( $room_id );
-               $room->_handle_event_initial( $_ ) for @$state;
-
-               $room->maybe_invoke_event( on_synced_state => );
-            }
-            elsif( $membership eq "invite" ) {
-               $self->maybe_invoke_event( on_invite => $_ );
-            }
-            else {
-               $self->log( "TODO: imsync returned a room in membership state $membership" );
-            }
-         }
-
-         # Now push use presence messages
-         foreach ( @{ $sync->{presence} } ) {
-            $self->_incoming_event( $_ );
-         }
-
-         $self->start_longpoll( start => $sync->{end} );
-         Future->done;
-      });
-      $self->adopt_future( $f );
+         return $self->await_synced->done;
+      })->on_fail( sub { undef $self->{start_f} });
    };
 }
 
-=head2 $matrix->stop
+=head2 stop
+
+   $matrix->stop
 
 Stops the event stream. After calling this you will need to use C<start> again
 to continue receiving events.
@@ -529,7 +638,7 @@ sub stop
 {
    my $self = shift;
 
-   ( delete $self->{start_f} )->cancel;
+   ( delete $self->{start_f} )->cancel if $self->{start_f};
    $self->stop_longpoll;
 }
 
@@ -541,24 +650,24 @@ sub start_longpoll
    my %args = @_;
 
    $self->stop_longpoll;
-   $self->{longpoll_last_token} = $args{start} // "END";
+   $self->{longpoll_last_token} = $args{since};
 
-   $self->{longpoll_f} = repeat {
+   my $f = $self->{longpoll_f} = repeat {
       my $last_token = $self->{longpoll_last_token};
 
-      my $uri = $self->_uri_for_path( "/events",
-         $last_token ? ( from => $last_token ) : (),
-         timeout => LONGPOLL_SECONDS * 1000, # msec
-      );
-
       Future->wait_any(
-         $self->loop->timeout_future( after => LONGPOLL_SECONDS + 5 ),
+         $self->{make_delay}->( $self->{longpoll_timeout} + 5 )
+            ->else_fail( "Longpoll timed out" ),
 
-         $self->{ua}->GET( $uri )->then( sub {
-            my ( $response ) = @_;
-            my $data = decode_json( $response->content );
-            $self->_incoming_event( $_ ) foreach @{ $data->{chunk} };
-            $self->{longpoll_last_token} = $data->{end};
+         $self->_do_GET_json( "/sync",
+            since   => $last_token,
+            timeout => $self->{longpoll_timeout} * 1000, # msec
+         )->then( sub {
+            my ( $sync ) = @_;
+
+            $self->_incoming_sync( $sync );
+
+            $self->{longpoll_last_token} = $sync->{next_batch};
 
             Future->done();
          }),
@@ -566,9 +675,16 @@ sub start_longpoll
          my ( $failure ) = @_;
          warn "Longpoll failed - $failure\n";
 
-         $self->loop->delay_future( after => 3 )
+         $self->{make_delay}->( 3 )
       });
-   } while => sub { 1 };
+   } while => sub { !shift->failure };
+
+   # Don't ->adopt_future this one as it makes it hard to grab to cancel it
+   # again, but apply the same on_fail => invoke_error logic
+   $f->on_fail( $self->_capture_weakself( sub {
+      my $self = shift;
+      $self->invoke_error( @_ );
+   }));
 }
 
 sub stop_longpoll
@@ -589,11 +705,12 @@ sub _get_or_make_user
 sub _make_room
 {
    my $self = shift;
-   my ( $room_id, @args ) = @_;
+   my ( $room_id ) = @_;
 
    $self->{rooms_by_id}{$room_id} and
       croak "Already have a room with ID '$room_id'";
 
+   my @args;
    foreach (qw( message member )) {
       push @args, "on_$_" => $self->{"on_room_$_"} if $self->{"on_room_$_"};
    }
@@ -619,13 +736,15 @@ sub make_room
 sub _get_or_make_room
 {
    my $self = shift;
-   my ( $room_id, @args ) = @_;
+   my ( $room_id ) = @_;
 
    return $self->{rooms_by_id}{$room_id} //
-      $self->_make_room( $room_id, @args );
+      $self->_make_room( $room_id );
 }
 
-=head2 $user = $matrix->myself
+=head2 myself
+
+   $user = $matrix->myself
 
 Returns the user object representing the connected user.
 
@@ -637,7 +756,9 @@ sub myself
    return $self->_get_or_make_user( $self->{user_id} );
 }
 
-=head2 $user = $matrix->user( $user_id )
+=head2 user
+
+   $user = $matrix->user( $user_id )
 
 Returns the user object representing a user of the given ID, if defined, or
 C<undef>.
@@ -651,26 +772,27 @@ sub user
    return $self->{users_by_id}{$user_id};
 }
 
-sub _incoming_event
+sub _incoming_sync
 {
    my $self = shift;
-   my ( $event ) = @_;
+   my ( $sync ) = @_;
 
-   my @type_parts = split m/\./, $event->{type};
-   my @subtype_args;
+   foreach my $category (qw( invite join leave )) {
+      my $rooms = $sync->{rooms}{$category} or next;
+      foreach my $room_id ( keys %$rooms ) {
+         my $roomsync = $rooms->{$room_id};
 
-   while( @type_parts ) {
-      if( my $handler = $self->can( "_handle_event_" . join "_", @type_parts ) ) {
-         $handler->( $self, @subtype_args, $event );
-         return;
+         my $room = $self->_get_or_make_room( $room_id );
+
+         $room->${\"_incoming_sync_$category"}( $roomsync );
       }
-
-      unshift @subtype_args, pop @type_parts;
    }
 
-   $self->maybe_invoke_event(
-      on_unknown_event => $event
-   ) or $self->log( "  incoming event=".pp($event) );
+   foreach my $event ( @{ $sync->{presence}{events} } ) {
+      $self->_handle_event_m_presence( $event );
+   }
+
+   # TODO: account_data
 }
 
 sub _on_self_leave
@@ -683,9 +805,13 @@ sub _on_self_leave
    delete $self->{rooms_by_id}{$room->room_id};
 }
 
-=head2 $name = $matrix->get_displayname->get
+=head2 get_displayname
 
-=head2 $matrix->set_displayname( $name )->get
+=head2 set_displayname
+
+   $name = $matrix->get_displayname->get
+
+   $matrix->set_displayname( $name )->get
 
 Accessor and mutator for the user account's "display name" profile field.
 
@@ -715,9 +841,13 @@ sub set_displayname
    );
 }
 
-=head2 ( $presence, $msg ) = $matrix->get_presence->get
+=head2 get_presence
 
-=head2 $matrix->set_presence( $presence, $msg )->get
+=head2 set_presence
+
+   ( $presence, $msg ) = $matrix->get_presence->get
+
+   $matrix->set_presence( $presence, $msg )->get
 
 Accessor and mutator for the user's current presence state and optional status
 message string.
@@ -788,7 +918,9 @@ sub drop_presence
    );
 }
 
-=head2 ( $room, $room_alias ) = $matrix->create_room( $alias_localpart )->get
+=head2 create_room
+
+   ( $room, $room_alias ) = $matrix->create_room( $alias_localpart )->get
 
 Requests the creation of a new room and associates a new alias with the given
 localpart on the server. The returned C<Future> will return an instance of
@@ -815,7 +947,9 @@ sub create_room
    });
 }
 
-=head2 $room = $matrix->join_room( $room_alias_or_id )->get
+=head2 join_room
+
+   $room = $matrix->join_room( $room_alias_or_id )->get
 
 Requests to join an existing room with the given alias name or plain room ID.
 If this room is already known by the C<$matrix> object, this method simply
@@ -828,29 +962,16 @@ sub join_room
    my $self = shift;
    my ( $room_alias ) = @_;
 
-   my $f;
-   if( $room_alias =~ m/^#/ ) {
-      $f = $self->_do_POST_json( "/join/$room_alias", {} )->then( sub {
-         my ( $content ) = @_;
-         Future->done( $content->{room_id} );
-      });
-   }
-   elsif( $room_alias =~ m/^!/ ) {
-      # Internal room ID directly
-      $f = $self->_do_PUT_json( "/rooms/$room_alias/state/m.room.member/$self->{user_id}", {
-         membership => "join",
-      } )->then_done( $room_alias );
-   }
+   $self->_do_POST_json( "/join/$room_alias", {} )->then( sub {
+      my ( $content ) = @_;
+      my $room_id = $content->{room_id};
 
-   $f->then( sub {
-      my ( $room_id ) = @_;
       if( my $room = $self->{rooms_by_id}{$room_id} ) {
          return Future->done( $room );
       }
       else {
          my $room = $self->_make_room( $room_id );
-         $room->initial_sync
-            ->then_done( $room );
+         return $room->await_synced->then_done( $room );
       }
    });
 }
@@ -866,9 +987,13 @@ sub room_list
       });
 }
 
-=head2 $matrix->add_alias( $alias, $room_id )->get
+=head2 add_alias
 
-=head2 $matrix->delete_alias( $alias )->get
+=head2 delete_alias
+
+   $matrix->add_alias( $alias, $room_id )->get
+
+   $matrix->delete_alias( $alias )->get
 
 Performs a directory server request to create the given room alias name, to
 point at the room ID, or to remove it again.
@@ -899,6 +1024,69 @@ sub delete_alias
       ->then_done();
 }
 
+=head2 upload
+
+   $content_uri = $matrix->upload( %params )->get
+
+Performs a post to the server's media content repository, to upload a new
+piece of content, returning the content URI that points to it.
+
+The content can be specified in any of three ways, with the following three
+mutually-exclusive arguments:
+
+=over 4
+
+=item content => STRING
+
+Gives the content directly as an immediate scalar value.
+
+=item file => STRING
+
+Gives the path to a readable file on the filesystem containing the content.
+
+=item fh => IO
+
+Gives an opened IO handle the content can be read from.
+
+=back
+
+The following additional arguments are also recognised:
+
+=over 4
+
+=item content_type => STRING
+
+Gives the MIME type of the content data.
+
+=item content_length => INT
+
+Optional. If the content is being delivered from an opened filehandle (via the
+C<fh> argument), this gives the total length in bytes. This is required in
+cases such as reading from pipes, when the length of the content isn't
+immediately available such as by C<stat()>ing the filehandle.
+
+=back
+
+=cut
+
+sub upload
+{
+   my $self = shift;
+   my %params = @_;
+
+   defined $params{content_type} or
+      croak "Require 'content_type'";
+
+   defined $params{content} or defined $params{file} or defined $params{fh} or
+      croak "Require 'content', 'file' or 'fh'";
+
+   # This one takes ~full URL paths
+   $self->_do_POST_file( "/media/v1/upload", %params )->then( sub {
+      my ( $content, $response ) = @_;
+      Future->done( $content->{content_uri} );
+   });
+}
+
 ## Incoming events
 
 sub _handle_event_m_presence
@@ -907,7 +1095,7 @@ sub _handle_event_m_presence
    my ( $event ) = @_;
    my $content = $event->{content};
 
-   my $user = $self->_get_or_make_user( $content->{user_id} );
+   my $user = $self->_get_or_make_user( $event->{sender} );
 
    my %changes;
    foreach (qw( presence displayname )) {
@@ -934,63 +1122,6 @@ sub _handle_event_m_presence
    }
 }
 
-sub _handle_event_m_room
-{
-   my $self = shift;
-   my $event = pop;
-   my @type_parts = @_;
-
-   # Room messages for existing rooms
-   my $handler;
-   if( $handler = $self->{rooms_by_id}{$event->{room_id}} ) {
-      # OK
-   }
-   elsif( $event->{state_key} eq $self->{user_id} ) {
-      $handler = $self;
-   }
-   else {
-      $self->log( "TODO: Room event on unknown room ID $event->{room_id} not about myself" );
-      # Ignore it for now
-      return;
-   }
-
-   my @subtype_parts;
-   while( @type_parts ) {
-      if( my $code = $handler->can( "_handle_roomevent_" . join "_", @type_parts, "forward" ) ) {
-         $code->( $handler, @subtype_parts, $event );
-         return;
-      }
-
-      unshift @subtype_parts, pop @type_parts;
-   }
-
-   $self->log( "Unhandled room event " . join( "_", @subtype_parts ) . "\n" .
-      pp( $event ) );
-}
-
-sub _handle_roomevent_member_forward
-{
-   my $self = shift;
-   my $event = pop;
-
-   my $content = $event->{content};
-   my $membership = $content->{membership};
-
-   if( $membership eq "join" ) {
-      my $room = $self->_get_or_make_room( $event->{room_id} );
-      $self->adopt_future(
-         # TODO: "members" isn't enough. We want other config too...
-         $room->initial_sync
-      );
-   }
-   elsif( $membership eq "invite" ) {
-      $self->maybe_invoke_event( on_invite => $event );
-   }
-   else {
-      $self->log( "Unhandled selfroom event member membership=$membership" );
-   }
-}
-
 =head1 USER STRUCTURES
 
 Parameters documented as C<$user> receive a user struct, which supports the
