@@ -1,7 +1,7 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2014 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2014-2016 -- leonerd@leonerd.org.uk
 
 package Net::Async::Matrix::Room;
 
@@ -11,18 +11,24 @@ use warnings;
 # Not really a Notifier but we like the ->maybe_invoke_event style
 use base qw( IO::Async::Notifier );
 
-our $VERSION = '0.11_002';
+our $VERSION = '0.18_001';
 $VERSION = eval $VERSION;
 
 use Carp;
 
 use Future;
+use Future::Utils qw( repeat );
 
 use List::Util qw( pairmap );
-use Struct::Dumb;
 use Time::HiRes qw( time );
 
-struct Member => [qw( user displayname membership )];
+use Net::Async::Matrix::Room::State;
+# TEMPORARY hack
+*Member = \&Net::Async::Matrix::Room::State::Member;
+
+use Data::Dump 'pp';
+
+use constant TYPING_RESEND_SECONDS => 30;
 
 =head1 NAME
 
@@ -104,6 +110,17 @@ C<$member> object will already be in the new state. C<%changes> will be a
 key/value list of state fields names that were changed, and references to
 2-element ARRAYs containing the old and new values for this field.
 
+=head2 on_typing $member, $is_typing
+
+Invoked on receipt of a typing notification change, when the given member
+either starts or stops typing.
+
+=head2 on_members_typing @members
+
+Invoked on receipt of a typing notification change to give the full set of
+currently-typing members. This is invoked after the individual C<on_typing>
+events.
+
 =cut
 
 sub _init
@@ -115,14 +132,11 @@ sub _init
    $self->{matrix}  = delete $params->{matrix};
    $self->{room_id} = delete $params->{room_id};
 
-   $self->{state} = {};
+   # Server gives us entire sets of typing user_ids all at once. We have to
+   # remember state
+   $self->{typing_members} = {};
 
-   $self->{members_by_userid} = {};
-   # Power levels can exist for users who aren't in the room. So store them
-   # separately, rather than on the member objects themselves
-   $self->{powerlevels} = {};
-
-   $self->{aliases_by_hs} = {};
+   $self->{live_state} = Net::Async::Matrix::Room::State->new( $self );
 }
 
 sub configure
@@ -131,7 +145,8 @@ sub configure
    my %params = @_;
 
    foreach (qw( on_message on_back_message on_membership on_back_membership
-         on_presence on_synced_state on_state_changed on_back_state_changed )) {
+         on_presence on_synced_state on_state_changed on_back_state_changed
+         on_typing on_members_typing )) {
       $self->{$_} = delete $params{$_} if exists $params{$_};
    }
 
@@ -142,6 +157,50 @@ sub configure
 
 =cut
 
+# FUNCTION
+sub _delete_null_changes
+{
+   my ( $changes ) = @_;
+
+   foreach ( keys %$changes ) {
+      my ( $old, $new ) = @{ $changes->{$_} };
+
+      delete $changes->{$_} if !defined $old and !defined $new or
+                               defined $old and defined $new and $old eq $new;
+   }
+}
+
+# FUNCTION
+sub _pushdown_changes
+{
+   my ( $ch ) = @_;
+   my ( $oldhash, $newhash ) = @$ch;
+
+   my %changes;
+
+   foreach ( keys %$oldhash ) {
+      my $old = $oldhash->{$_};
+      if( !exists $newhash->{$_} ) {
+         $changes{$_} = [ $old, undef ] if defined $old;
+         next;
+      }
+
+      my $new = $newhash->{$_};
+
+      $changes{$_} = [ $old, $new ] unless !defined $old and !defined $new or
+                                           defined $old and defined $new and $old eq $new;
+   }
+
+   foreach ( keys %$newhash ) {
+      my $new = $newhash->{$_};
+      next if exists $oldhash->{$_};
+
+      $changes{$_} = [ undef, $new ] if defined $new;
+   }
+
+   return keys %changes ? \%changes : undef;
+}
+
 sub _do_GET_json
 {
    my $self = shift;
@@ -166,55 +225,151 @@ sub _do_POST_json
    $self->{matrix}->_do_POST_json( "/rooms/$self->{room_id}" . $path, $content );
 }
 
-sub initial_sync
+=head2 await_synced
+
+   $f = $room->await_synced
+
+Returns a L<Future> stored within the room that will complete (with no value)
+once the room initial state sync has been completed. This completes just
+I<before> the C<on_synced_state> event.
+
+=cut
+
+sub _reset_for_sync
 {
    my $self = shift;
 
-   # There is not actually an 'initialSync' API for individual rooms, yet. See
-   #   https://matrix.org/jira/browse/SYN-55
-
-   $self->{initial_sync} ||= $self->_do_GET_json( "/state" )->then( sub {
-      my ( $events ) = @_;
-
-      $self->_handle_event_initial( $_ ) for @$events;
-
-      $self->maybe_invoke_event( on_synced_state => );
-      Future->done;
-   });
+   undef $self->{synced_future};
 }
 
-sub _handle_event_initial
+sub _incoming_sync_invite
 {
    my $self = shift;
-   my ( $event ) = @_;
+   my ( $sync ) = @_;
+   warn "TODO handle incoming sync data in invite state";
+}
 
-   $event->{type} =~ m/^m\.room\.(.*)$/ or return;
-   my $method = "_handle_roomevent_" . join( "_", split m/\./, $1 ) . "_initial";
+sub _incoming_sync_join
+{
+   my $self = shift;
+   my ( $sync ) = @_;
+
+   my $initial = not $self->await_synced->is_done;
+
+   # Toplevel fields for now I'm ignoring
+   #   account_data
+   #   unread_notifications
+
+   my $live_state = $self->live_state;
+
+   if( $sync->{state} and $sync->{state}{events} and @{ $sync->{state}{events} } ) {
+      foreach my $event ( @{ $sync->{state}{events} } ) {
+         $live_state->handle_event( $event );
+      }
+   }
+
+   foreach my $event ( @{ $sync->{timeline}{events} } ) {
+      if( defined $event->{state_key} ) {
+         my $old_event = $live_state->get_event( $event->{type}, $event->{state_key} );
+         $live_state->handle_event( $event );
+         $self->_handle_state_event( $old_event, $event, $live_state );
+      }
+      else {
+         $self->_handle_event( forward => $event );
+      }
+   }
+
+   foreach my $event ( @{ $sync->{ephemeral}{events} } ) {
+      $self->_handle_event( ephemeral => $event );
+   }
+
+   if( $initial ) {
+      $self->await_synced->done;
+      $self->maybe_invoke_event( on_synced_state => );
+   }
+}
+
+sub _incoming_sync_leave
+{
+   my $self = shift;
+   my ( $sync ) = @_;
+   # don't care for now
+}
+
+sub await_synced
+{
+   my $self = shift;
+   return $self->{synced_future} //= $self->loop->new_future;
+}
+
+=head2 live_state
+
+   $state = $room->live_state
+
+Returns a L<Net::Async::Matrix::Room::State> instance representing the current
+live-tracking state of the room.
+
+This instance will mutate and change as new state events are received.
+
+=cut
+
+sub live_state
+{
+   my $self = shift;
+   return $self->{live_state};
+}
+
+sub _handle_state_event
+{
+   my $self = shift;
+   my ( $old_event, $new_event, $state ) = @_;
+
+   my $old_content = $old_event->{content};
+   my $new_content = $new_event->{content};
+
+   my %changes;
+   $changes{$_}->[0] = $old_content->{$_} for keys %$old_content;
+   $changes{$_}->[1] = $new_content->{$_} for keys %$new_content;
+
+   $_->[1] //= undef for values %changes; # Ensure deleted key values become undef
+
+   _delete_null_changes \%changes;
+
+   my $member = $state->member( $new_event->{sender} );
+
+   my $type = $new_event->{type};
+
+   $type =~ m/^m\.room\.(.*)$/;
+   my $method = $1 ? "_handle_state_event_" . join( "_", split m/\./, $1 ) : undef;
+
+   if( $method and my $code = $self->can( $method ) ) {
+      $self->$code( $member, $new_event, $state, %changes );
+   }
+   else {
+      $self->maybe_invoke_event( on_state_changed =>
+         $member, $new_event, %changes
+      );
+   }
+}
+
+sub _handle_event
+{
+   my $self = shift;
+   my ( $direction, $event ) = @_;
+
+   $event->{type} =~ m/^(m\.room\.)?(.*)$/ or return;
+
+   my $base = $1 ? "_handle_roomevent_" : "_handle_event_";
+   my $method = $base . join( "_", split( m/\./, $2 ), $direction );
 
    if( my $code = $self->can( $method ) ) {
       $code->( $self, $event );
    }
    else {
-      warn "TODO: initial room event $event->{type}\n";
+      warn "TODO: $direction event $event->{type}\n";
    }
 }
 
-sub _handle_state_forward
-{
-   my $self = shift;
-   my ( $field, $event ) = @_;
-
-   my $newvalue = $event->{content}{$field};
-
-   my $oldvalue = $self->{state}{$field};
-   $self->{state}{$field} = $newvalue;
-
-   $self->maybe_invoke_event( on_state_changed =>
-      $self->{members_by_userid}{$event->{user_id}}, $event,
-      $field => [ $oldvalue, $newvalue ]
-   );
-}
-
 sub _handle_state_backward
 {
    my $self = shift;
@@ -229,7 +384,9 @@ sub _handle_state_backward
    );
 }
 
-=head2 $id = $room->room_id
+=head2 room_id
+
+   $id = $room->room_id
 
 Returns the opaque room ID string for the room. Usually this would not be
 required, except for long-term persistence uniqueness purposes, or for
@@ -243,26 +400,14 @@ sub room_id
    return $self->{room_id};
 }
 
-=head2 $name = $room->name
+=head2 name
+
+   $name = $room->name
 
 Returns the room name, if defined, otherwise the opaque room ID.
 
 =cut
 
-sub _handle_roomevent_name_initial
-{
-   my $self = shift;
-   my ( $event ) = @_;
-   $self->{state}{name} = $event->{content}{name};
-}
-
-sub _handle_roomevent_name_forward
-{
-   my $self = shift;
-   my ( $event ) = @_;
-   $self->_handle_state_forward( name => $event );
-}
-
 sub _handle_roomevent_name_backward
 {
    my $self = shift;
@@ -273,10 +418,12 @@ sub _handle_roomevent_name_backward
 sub name
 {
    my $self = shift;
-   return $self->{state}{name} || $self->room_id;
+   return $self->live_state->name || $self->room_id;
 }
 
-=head2 $room->set_name( $name )->get
+=head2 set_name
+
+   $room->set_name( $name )->get
 
 Requests to set a new room name.
 
@@ -291,7 +438,9 @@ sub set_name
       ->then_done();
 }
 
-=head2 @aliases = $room->aliases
+=head2 aliases
+
+   @aliases = $room->aliases
 
 Returns a list of all the known room alias names taken from the
 C<m.room.alias> events. Note that these are simply names I<claimed> to have
@@ -301,35 +450,21 @@ their values.
 
 =cut
 
-sub _handle_roomevent_aliases_initial
+sub _handle_state_event_aliases
 {
    my $self = shift;
-   my ( $event ) = @_;
+   my ( $member, $event, $state, %changes ) = @_;
 
    my $homeserver = $event->{state_key};
 
-   $self->{aliases_by_hs}{$homeserver} = [ @{ $event->{content}{aliases} } ];
-}
+   my @others = map { $_->{content}{aliases} }
+                grep { $_->{state_key} ne $homeserver }
+                values %{ $state->get_events( "m.room.aliases" ) };
 
-sub _handle_roomevent_aliases_forward
-{
-   my $self = shift;
-   my ( $event ) = @_;
-
-   my $homeserver = $event->{state_key};
-
-   my $new = $event->{content}{aliases} // [];
-   my $old = $event->{prev_content}{aliases} // [];
-
-   $self->{aliases_by_hs}{$homeserver} = [ @$new ];
-
-   my @others = map { @{ $self->{aliases_by_hs}{$_} } }
-                grep { $_ ne $homeserver }
-                keys %{ $self->{aliases_by_hs} };
+   $changes{aliases}[2] = \@others;
 
    $self->maybe_invoke_event( on_state_changed =>
-      $self->{members_by_userid}{$event->{user_id}}, $event,
-      aliases => [ $old, $new, \@others ]
+      $member, $event, %changes
    );
 }
 
@@ -358,11 +493,12 @@ sub _handle_roomevent_aliases_backward
 sub aliases
 {
    my $self = shift;
-
-   return map { @$_ } values %{ $self->{aliases_by_hs} };
+   return $self->live_state->aliases;
 }
 
-=head2 $rule = $room->join_rule
+=head2 join_rule
+
+   $rule = $room->join_rule
 
 Returns the current C<join_rule> for the room; a string giving the type of
 access new members may get:
@@ -389,20 +525,6 @@ No new users may join the room
 
 =cut
 
-sub _handle_roomevent_join_rules_initial
-{
-   my $self = shift;
-   my ( $event ) = @_;
-   $self->{state}{join_rule} = $event->{content}{join_rule};
-}
-
-sub _handle_roomevent_join_rules_forward
-{
-   my $self = shift;
-   my ( $event ) = @_;
-   $self->_handle_state_forward( join_rule => $event );
-}
-
 sub _handle_roomevent_join_rules_backward
 {
    my $self = shift;
@@ -413,29 +535,17 @@ sub _handle_roomevent_join_rules_backward
 sub join_rule
 {
    my $self = shift;
-   return $self->{state}{join_rule};
+   return $self->live_state->join_rule;
 }
 
-=head2 $topic = $room->topic
+=head2 topic
+
+   $topic = $room->topic
 
 Returns the room topic, if defined
 
 =cut
 
-sub _handle_roomevent_topic_initial
-{
-   my $self = shift;
-   my ( $event ) = @_;
-   $self->{state}{topic} = $event->{content}{topic};
-}
-
-sub _handle_roomevent_topic_forward
-{
-   my $self = shift;
-   my ( $event ) = @_;
-   $self->_handle_state_forward( topic => $event );
-}
-
 sub _handle_roomevent_topic_backward
 {
    my $self = shift;
@@ -446,10 +556,12 @@ sub _handle_roomevent_topic_backward
 sub topic
 {
    my $self = shift;
-   return $self->{state}{topic};
+   return $self->live_state->topic;
 }
 
-=head2 $room->set_topic( $topic )->get
+=head2 set_topic
+
+   $room->set_topic( $topic )->get
 
 Requests to set a new room topic.
 
@@ -464,7 +576,9 @@ sub set_topic
       ->then_done();
 }
 
-=head2 %levels = $room->levels
+=head2 levels
+
+   %levels = $room->levels
 
 Returns a key/value list of the room levels; that is, the member power level
 required to perform each of the named actions.
@@ -501,7 +615,7 @@ sub _handle_generic_level
             if !defined $oldlevels->{$_} or $oldlevels->{$_} != $newlevels->{$_};
       }
 
-      my $member = $self->{members_by_userid}{$event->{user_id}};
+      my $member = $self->member( $event->{sender} );
       $self->maybe_invoke_event( on_state_changed =>
          $member, $event, %changes
       );
@@ -523,37 +637,15 @@ sub _handle_generic_level
    }
 }
 
-{
-   foreach my $phase (qw( initial forward backward )) {
-      no strict 'refs';
-
-      *{"_handle_roomevent_ops_levels_${phase}"} = sub {
-         shift->${\"_handle_generic_level"}( $phase, ops =>
-            { map {; "${_}_level", $_ } qw( ban kick redact ) }, @_
-         );
-      };
-
-      *{"_handle_roomevent_send_event_level_${phase}"} = sub {
-         shift->${\"_handle_generic_level"}( $phase, send_event =>
-            { level => "send_event" }, @_
-         );
-      };
-
-      *{"_handle_roomevent_add_state_level_${phase}"} = sub {
-         shift->${\"_handle_generic_level"}( $phase, add_state =>
-            { level => "add_state" }, @_
-         );
-      };
-   }
-}
-
 sub levels
 {
    my $self = shift;
    return %{ $self->{levels} };
 }
 
-=head2 $room->change_levels( %levels )->get
+=head2 change_levels
+
+   $room->change_levels( %levels )->get
 
 Performs a room levels change, submitting new values for the given keys while
 leaving other keys unchanged.
@@ -592,7 +684,9 @@ sub change_levels
    )->then_done();
 }
 
-=head2 @members = $room->members
+=head2 members
+
+   @members = $room->members
 
 Returns a list of member structs containing the currently known members of the
 room, in no particular order. This list will include users who are not yet
@@ -600,98 +694,55 @@ members of the room, but simply have been invited.
 
 =cut
 
-sub _handle_roomevent_member_initial
-{
-   my $self = shift;
-   my ( $event ) = @_;
-
-   my $user_id = $event->{state_key}; # == user the change applies to
-   my $content = $event->{content};
-
-   warn "ARGH: Room '$self->{room_id}' already has a member '$user_id'\n" and return
-      if $self->{members_by_userid}{$user_id};
-
-   my $user = $self->{matrix}->_get_or_make_user( $user_id );
-
-   $self->{members_by_userid}{$user_id} = Member(
-      $user, $content->{displayname}, $content->{membership} );
-}
-
-sub _handle_roomevent_member_forward
-{
-   my $self = shift;
-   my ( $event ) = @_;
-
-   $self->_handle_roomevent_member( on_membership => $event,
-      $self->{members_by_userid}, $event->{prev_content}, $event->{content} );
-
-   my $matrix = $self->{matrix};
-   if( $event->{content}{membership} eq "leave" and $event->{state_key} eq $matrix->{user_id} ) {
-      $matrix->_on_self_leave( $self );
-   }
-}
-
 sub _handle_roomevent_member_backward
 {
    my $self = shift;
    my ( $event ) = @_;
 
-   $self->_handle_roomevent_member( on_back_membership => $event,
-      $self->{back_members_by_userid}, $event->{content}, $event->{prev_content} );
+   # $self->_handle_roomevent_member( on_back_membership => $event,
+   #    $self->{back_members_by_userid}, $event->{content}, $event->{prev_content} );
 }
 
-sub _handle_roomevent_member
+sub _handle_state_event_member
 {
    my $self = shift;
-   my ( $name, $event, $members, $old, $new ) = @_;
+   my ( $member, $event, $state, %changes ) = @_;
 
-   # Currently, the server "deletes" users from the membership by setting
+   # Currently, the server "deletes" users from the room by setting their
    # membership to "leave". It's neater if we consider an empty content in
    # that case.
-   $_ and $_->{membership} and $_->{membership} eq "leave" and undef $_
-      for $old, $new;
+   foreach my $idx ( 0, 1 ) {
+      next unless ( $changes{membership}[$idx] // "" ) eq "leave";
 
-   $_ and not keys %$_ and undef $_
-      for $old, $new;
+      undef $changes{$_}[$idx] for keys %changes;
+   }
 
    my $user_id = $event->{state_key}; # == user the change applies to
 
-   my $target_member;
-   if( $old ) {
-      $target_member = $members->{$user_id} or
-         warn "ARGH: roomevent_member with unknown user id '$user_id'" and return;
-   }
-   else {
-      my $user = $self->{matrix}->_get_or_make_user( $user_id );
-      $target_member = $members->{$user_id} ||=
-         Member( $user, undef, undef );
-   }
+   my $target_member = $state->member( $user_id ) or
+      warn "ARGH: roomevent_member with unknown user id '$user_id'" and return;
 
-   my %changes;
-   foreach (qw( membership displayname )) {
-      next if !defined $old->{$_} and !defined $new->{$_};
-      next if defined $old->{$_} and defined $new->{$_} and $old->{$_} eq $new->{$_};
+   _delete_null_changes \%changes;
 
-      $changes{$_} = [ $old->{$_}, $new->{$_} ];
-      $target_member->$_ = $new->{$_};
-   }
-
-   my $member = $members->{$event->{user_id}}; # == the user making the change
-
-   $self->maybe_invoke_event( $name => $member, $event, $target_member, %changes );
-
-   if( !$new ) {
-      delete $members->{$user_id};
-   }
+   $self->maybe_invoke_event( on_membership => $member, $event, $target_member, %changes );
 }
 
 sub members
 {
    my $self = shift;
-   return values %{ $self->{members_by_userid} };
+   return $self->live_state->members;
 }
 
-=head2 @members = $room->joined_members
+sub member
+{
+   my $self = shift;
+   my ( $user_id ) = @_;
+   return $self->live_state->member( $user_id );
+}
+
+=head2 joined_members
+
+   @members = $room->joined_members
 
 Returns the subset of C<all_members> who actually in the C<"join"> state -
 i.e. are not invitees, or have left.
@@ -701,84 +752,53 @@ i.e. are not invitees, or have left.
 sub joined_members
 {
    my $self = shift;
-   return grep { $_->membership eq "join" } $self->members;
+   return grep { ( $_->membership // "" ) eq "join" } $self->members;
 }
 
-=head2 $level = $room->member_level( $user_id )
+=head2 member_level
+
+   $level = $room->member_level( $user_id )
 
 Returns the current cached value for the power level of the given user ID, or
 the default value if no specific value exists for the given ID.
 
 =cut
 
-sub _handle_roomevent_power_levels_initial
-{
-   my $self = shift;
-   my ( $event ) = @_;
-
-   my $levels = $event->{content};
-   $self->{powerlevels} = { %$levels };
-}
-
-sub _handle_roomevent_power_levels_forward
-{
-   my $self = shift;
-   my ( $event ) = @_;
-
-   my $levels = $event->{content};
-   $self->{powerlevels} = { %$levels };
-
-   $self->_handle_roomevent_power_levels( on_membership =>
-      $event, $self->{members_by_userid}, $event->{prev_content}, $levels
-   );
-}
-
 sub _handle_roomevent_power_levels_backward
 {
    my $self = shift;
    my ( $event ) = @_;
 
-   $self->_handle_roomevent_power_levels( on_back_membership =>
-      $event, $self->{back_members_by_userid}, $event->{content}, $event->{prev_content}
-   );
+   # $self->_handle_roomevent_power_levels( on_back_membership =>
+   #    $event, $self->{back_members_by_userid}, $event->{content}, $event->{prev_content}
+   # );
 }
 
-sub _handle_roomevent_power_levels
+sub _handle_state_event_power_levels
 {
    my $self = shift;
-   my ( $name, $event, $members, $old, $new ) = @_;
+   my ( $member, $event, $state, %changes ) = @_;
 
-   $old = $old->{users};
-   $new = $new->{users};
+   # Before we go any further we should also clean up null changes in 'users'
+   # and 'events' hashes by pushing the 'old+new' diff ARRAYrefs down into the
+   # hashes
+   $_ and $_ = _pushdown_changes $_ for $changes{users}, $changes{events};
 
-   my $change_member = $members->{$event->{user_id}};
+   if( my $users = $changes{users} ) {
+      # TODO: handle default changes
+      my $default = $event->{content}{user_default};
 
-   foreach my $user_id ( keys %$new ) {
-      next if $user_id eq "default";
+      foreach my $user_id ( keys %$users ) {
+         my $target = $state->member( $user_id ) or next;
+         my ( $oldlevel, $newlevel ) = @{ $users->{$user_id} };
 
-      my $newlevel = $new->{$user_id} // $new->{default};
-      my $oldlevel = $old->{$user_id} // $old->{default};
-      next if defined $newlevel and defined $oldlevel and $newlevel == $oldlevel;
+         $oldlevel //= $default;
+         $newlevel //= $default;
 
-      my $member = $members->{$user_id} or next;
-
-      $self->maybe_invoke_event( $name =>
-         $change_member, $event, $member, level => [ $oldlevel, $newlevel ]
-      );
-   }
-
-   foreach my $user_id ( keys %$old ) {
-      next if exists $new->{$user_id};
-
-      my $newlevel = $new->{$user_id} // $new->{default};
-      my $oldlevel = $old->{$user_id} // $old->{default};
-      next if defined $newlevel and defined $oldlevel and $newlevel == $oldlevel;
-
-      my $member = $members->{$user_id} or next;
-
-      $self->maybe_invoke_event( $name =>
-         $change_member, $event, $member, level => [ $oldlevel, $newlevel ]
-      );
+         $self->maybe_invoke_event( on_membership =>
+            $member, $event, $target, level => [ $oldlevel, $newlevel ]
+         );
+      }
    }
 }
 
@@ -786,11 +806,12 @@ sub member_level
 {
    my $self = shift;
    my ( $user_id ) = @_;
-
-   return $self->{powerlevels}{users}{$user_id} // $self->{powerlevels}{users_default};
+   return $self->live_state->member_level( $user_id );
 }
 
-=head2 $room->change_member_levels( %levels )->get
+=head2 change_member_levels
+
+   $room->change_member_levels( %levels )->get
 
 Performs a member power level change, submitting new values for user IDs to
 the home server. As there is no server API to make individual mutations, this
@@ -832,7 +853,9 @@ sub change_member_levels
    )->then_done();
 }
 
-=head2 $room->leave->get
+=head2 leave
+
+   $room->leave->get
 
 Requests to leave the room. After this completes, the user will no longer be
 a member of the room.
@@ -845,7 +868,9 @@ sub leave
    $self->_do_POST_json( "/leave", {} );
 }
 
-=head2 $room->invite( $user_id )->get
+=head2 invite
+
+   $room->invite( $user_id )->get
 
 Sends an invitation for the user with the given User ID to join the room.
 
@@ -860,7 +885,9 @@ sub invite
       ->then_done();
 }
 
-=head2 $room->send_message( %args )->get
+=head2 send_message
+
+   $event_id = $room->send_message( %args )->get
 
 Sends a new message to the room. Requires a C<type> named argument giving the
 message type. Depending on the type, further keys will be required that
@@ -868,21 +895,25 @@ specify the message contents:
 
 =over 4
 
-=item text, emote
+=item m.text, m.emote, m.notice
 
 Require C<body>
 
-=item image, audio, video
+=item m.image, m.audio, m.video, m.file
 
 Require C<url>
 
-=item location
+=item m.location
 
 Require C<geo_uri>
 
 =back
 
-=head2 $room->send_message( $text )->get
+If an additional argument called C<txn_id> is provided, this is used as the
+transaction ID for the message, which is then sent as a C<PUT> request instead
+of a C<POST>.
+
+   $event_id = $room->send_message( $text )->get
 
 A convenient shortcut to sending an C<text> message with a body string and
 no additional content.
@@ -890,11 +921,13 @@ no additional content.
 =cut
 
 my %MSG_REQUIRED_FIELDS = (
-   'm.text'  => [qw( body )],
-   'm.emote' => [qw( body )],
-   'm.image' => [qw( url )],
-   'm.audio' => [qw( url )],
-   'm.video' => [qw( url )],
+   'm.text'     => [qw( body )],
+   'm.emote'    => [qw( body )],
+   'm.notice'   => [qw( body )],
+   'm.image'    => [qw( url )],
+   'm.audio'    => [qw( url )],
+   'm.video'    => [qw( url )],
+   'm.file'     => [qw( url )],
    'm.location' => [qw( geo_uri )],
 );
 
@@ -913,11 +946,23 @@ sub send_message
       $args{$_} or croak "'$type' messages require a '$_' field";
    }
 
-   $self->_do_POST_json( "/send/m.room.message", \%args )
-      ->then_done()
+   if( defined( my $txn_id = $args{txn_id} ) ) {
+      $self->_do_PUT_json( "/send/m.room.message/$txn_id", \%args )->then( sub {
+         my ( $response ) = @_;
+         Future->done( $response->{event_id} );
+      });
+   }
+   else {
+      $self->_do_POST_json( "/send/m.room.message", \%args )->then( sub {
+         my ( $response ) = @_;
+         Future->done( $response->{event_id} );
+      });
+   }
 }
 
-=head2 $room->paginate_messages( limit => $n )->get
+=head2 paginate_messages
+
+   $room->paginate_messages( limit => $n )->get
 
 Requests more messages of back-pagination history.
 
@@ -971,6 +1016,67 @@ sub paginate_messages
    $self->adopt_future( $f );
 }
 
+=head2 typing_start
+
+   $room->typing_start
+
+Sends a typing notification that the user is currently typing in this room.
+This notification will periodically be re-sent as required by the protocol
+until the C<typing_stop> method is called.
+
+=cut
+
+sub typing_start
+{
+   my $self = shift;
+
+   return if $self->{typing_timer};
+
+   my $user_id = $self->{matrix}->myself->user_id;
+
+   my $f = $self->{typing_timer} = repeat {
+      $self->_do_PUT_json( "/typing/$user_id", {
+         typing  => 1,
+         timeout => ( TYPING_RESEND_SECONDS + 5 ) * 1000, # msec
+      })->then( sub {
+         $self->{matrix}->{make_delay}->( TYPING_RESEND_SECONDS );
+      });
+   } while => sub { !shift->failure };
+
+   $f->on_fail( $self->_capture_weakself( sub {
+      my $self = shift;
+      $self->invoke_error( @_ );
+   }));
+}
+
+=head2 typing_stop
+
+   $room->typing_stop
+
+Sends a typing notification that the user is no longer typing in this room.
+This method also cancels the repeating re-send behaviour created by
+C<typing_start>.
+
+=cut
+
+sub typing_stop
+{
+   my $self = shift;
+
+   return unless my $f = $self->{typing_timer};
+
+   $f->cancel;
+   undef $self->{typing_timer};
+
+   my $user_id = $self->{matrix}->myself->user_id;
+
+   $self->adopt_future(
+      $self->_do_PUT_json( "/typing/$user_id", {
+         typing => 0,
+      })
+   );
+}
+
 sub _handle_roomevent_create_forward
 {
    my $self = shift;
@@ -993,8 +1099,8 @@ sub _handle_roomevent_message_forward
    my $self = shift;
    my ( $event ) = @_;
 
-   my $user_id = $event->{user_id};
-   my $member = $self->{members_by_userid}{$user_id} or
+   my $user_id = $event->{sender};
+   my $member = $self->member( $user_id ) or
       warn "TODO: Unknown member '$user_id' for forward message" and return;
 
    $self->maybe_invoke_event( on_message => $member, $event->{content}, $event );
@@ -1016,7 +1122,7 @@ sub _handle_event_m_presence
 {
    my $self = shift;
    my ( $user, %changes ) = @_;
-   my $member = $self->{members_by_userid}{$user->user_id} or return;
+   my $member = $self->member( $user->user_id ) or return;
 
    $changes{$_} and $member->$_ = $changes{$_}[1]
       for qw( displayname );
@@ -1024,6 +1130,33 @@ sub _handle_event_m_presence
    $self->maybe_invoke_event( on_presence => $member, %changes );
 }
 
+sub _handle_event_m_typing_ephemeral
+{
+   my $self = shift;
+   my ( $event ) = @_;
+
+   my $typing = $self->{typing_members};
+   my %not_typing = %$typing;
+
+   foreach my $user_id ( @{ $event->{content}{user_ids} } ) {
+      delete $not_typing{$user_id};
+      next if $typing->{$user_id};
+
+      $typing->{$user_id}++;
+      my $member = $self->member( $user_id ) or next;
+      $self->maybe_invoke_event( on_typing => $member, 1 );
+   }
+
+   foreach my $user_id ( keys %not_typing ) {
+      my $member = $self->member( $user_id ) or next;
+      $self->maybe_invoke_event( on_typing => $member, 0 );
+      delete $typing->{$user_id};
+   }
+
+   my @members = map { $self->member( $_ ) } keys %$typing;
+   $self->maybe_invoke_event( on_members_typing => grep { defined } @members );
+}
+
 =head1 MEMBERSHIP STRUCTURES
 
 Parameters documented as C<$member> receive a membership struct, which
